from datetime import datetime
from unittest.mock import Mock

from spetlrtools.testing import DataframeTestCase

from spetlr.delta import DeltaHandle


class DeltaHandleStreamTests(DataframeTestCase):
    def test_01_stream_starttime_datetime(self):
        DeltaHandle("test", stream_start=datetime(2020, month=1, day=1)).append(Mock())

    def test_02_stream_starttime_string(self):
        DeltaHandle("test", stream_start="2020-01-01").append(Mock())



import unittest

from spetlr.entry_points.generalized_task_entry_point import prepare_keyword_arguments


class TestArgumnetHandler(unittest.TestCase):
    def test_good_parameters(self):
        def f(a, b=1, c=1):
            pass

        kwargs = {"a": 4, "b": 5, "c": 6}
        result = prepare_keyword_arguments(f, kwargs)
        self.assertEqual(result, kwargs)

    def test_too_many_parameters(self):
        def f(a, b=1, c=1):
            pass

        kwargs = {"b": 5, "c": 6, "d": 4}
        result = prepare_keyword_arguments(f, kwargs)
        self.assertEqual(set(result.keys()), set("bc"))



import re
from typing import Dict, List

import sqlparse
import yaml
from sqlparse.sql import Statement

_magic_comment_start = "-- spetlr.configurator "


def _extract_comment_attributes(stmt: Statement) -> Dict:
    _magic_comment_start_re = re.compile(_magic_comment_start, re.IGNORECASE)
    collected_lines = []
    for token in _list_no_whitespace(stmt):
        if token.ttype not in sqlparse.tokens.Comment:
            continue
        val: str = token.value.strip()
        if not _magic_comment_start_re.match(val):
            continue
        collected_lines.append(val[len(_magic_comment_start) :])
    yaml_document = "\n".join(collected_lines)
    return yaml.load(yaml_document, Loader=yaml.FullLoader) or {}


def _list_no_whitespace(
    tokens: Statement,
) -> List[sqlparse.tokens._TokenType]:
    return [t for t in tokens if t.ttype != sqlparse.tokens.Whitespace]



from typing import Any, List, Union

from pyspark.sql import DataFrame
from pyspark.sql.types import StructType

from spetlr.cosmos.cosmos_base_server import CosmosBaseServer
from spetlr.exceptions import SpetlrException
from spetlr.tables.TableHandle import TableHandle


class CosmosHandleException(SpetlrException):
    pass


class CosmosHandleInvalidName(CosmosHandleException):
    pass


class CosmosHandleInvalidFormat(CosmosHandleException):
    pass


class CosmosHandle(TableHandle):
    def __init__(
        self,
        name: str,
        cosmos_db: CosmosBaseServer,
        rows_per_partition: int = None,
        schema: StructType = None,
    ):
        self._name = name
        self._cosmos_db = cosmos_db
        self._rows_per_partition = rows_per_partition
        self._schema = schema

    def read(self) -> DataFrame:
        return self._cosmos_db.read_table_by_name(
            table_name=self._name, schema=self._schema
        )

    def recreate(self):
        self._cosmos_db.recreate_container_by_name(self._name)

    def append(
        self,
        df: DataFrame,
        mergeSchema: bool = None,
    ) -> None:
        # ignore mergeSchema silently. Schema evolution works implicitly in cosmos
        self._cosmos_db.write_table_by_name(df, self._name, self._rows_per_partition)

    def truncate(self) -> None:
        self.recreate()

    def drop(self) -> None:
        self._cosmos_db.delete_container_by_name(self._name)

    def drop_and_delete(self) -> None:
        self.drop()

    def write_or_append(self, df: DataFrame, mode: str) -> None:
        if mode == "append":
            return self.append(df)
        else:
            raise ValueError(f"unsupported flag value of mode: {mode}")

    def overwrite(
        self, df: DataFrame, mergeSchema: bool = None, overwriteSchema: bool = None
    ) -> None:
        raise NotImplementedError("Method not supported in Cosmos")

    def upsert(self, df: DataFrame, join_cols: List[str]) -> Union[DataFrame, None]:
        raise NotImplementedError("Method not supported in Cosmos")

    def get_tablename(self) -> str:
        return self._name

    def delete_data(
        self, comparison_col: str, comparison_limit: Any, comparison_operator: str
    ) -> None:
        raise NotImplementedError("Method not yet implemented in Cosmos")



import time
import unittest
import uuid
from typing import List, Tuple

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.etl import Orchestrator
from spetlr.etl.extractors.stream_extractor import StreamExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.etl.loaders.stream_loader import StreamLoader
from spetlr.filehandle import FileHandle
from spetlr.functions import init_dbutils
from spetlr.spark import Spark
from spetlr.testutils.stop_test_streams import stop_test_streams
from tests.cluster.values import resourceName


@unittest.skipUnless(
    Spark.version() >= Spark.DATABRICKS_RUNTIME_10_4,
    f"Autoloader not available for Spark version {Spark.version()}",
)
class FileHandleTests(unittest.TestCase):
    sink_checkpoint_path: str = None
    path_unique_id = uuid.uuid4().hex
    avrosource_checkpoint_path = (
        f"/mnt/{resourceName()}/silver/{resourceName()}"
        f"/avrolocation/{path_unique_id}/_checkpoint_path_avro"
    )

    avro_source_path = (
        f"/mnt/{resourceName()}/silver/{resourceName()}/"
        f"avrolocation/{path_unique_id}/AvroSource"
    )

    @classmethod
    def setUpClass(cls) -> None:
        Configurator().clear_all_configurations()
        Configurator().set_debug()

        init_dbutils().fs.mkdirs(cls.avrosource_checkpoint_path)

        init_dbutils().fs.mkdirs(cls.avro_source_path)

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle.from_tc("MyDb").drop_cascade()

        init_dbutils().fs.rm(cls.avrosource_checkpoint_path, True)

        init_dbutils().fs.rm(cls.avro_source_path, True)

        stop_test_streams()

    def test_01_configure(self):
        tc = Configurator()
        tc.register(
            "MyDb", {"name": "TestDb{ID}", "path": "/mnt/spetlr/silver/testdb{ID}"}
        )
        # add avro source
        tc.register(
            "AvroSource",
            {
                "name": "AvroSource",
                "path": self.avro_source_path,
                "format": "avro",
                "partitioning": "ymd",
                "schema_location": self.avrosource_checkpoint_path,
            },
        )

        # Add sink table
        self.sink_checkpoint_path = (
            "/mnt/spetlr/silver/testdb{ID}/_checkpoint_path_avrosink"
        )

        # add eventhub sink
        tc.register(
            "AvroSink",
            {
                "name": "{MyDb}.AvroSink",
                "path": "{MyDb_path}/AvroSink",
                "format": "delta",
                "schema_location": self.sink_checkpoint_path,
                "await_termination": True,
                "query_name": "testquery{ID}",
            },
        )

        # test instantiation without error
        DbHandle.from_tc("MyDb")
        FileHandle.from_tc("AvroSource")
        DeltaHandle.from_tc("AvroSink")

    def test_02_read_avro(self):
        tc = Configurator()
        DbHandle.from_tc("MyDb").create()

        dh_sink = DeltaHandle.from_tc("AvroSink")
        Spark.get().sql(
            f"""
                            CREATE TABLE {dh_sink.get_tablename()}
                            (
                            id int,
                            name string,
                            _rescued_data string
                            )
                            USING DELTA
                            LOCATION '{tc.get("AvroSink","path")}'
                        """
        )

        self._add_avro_data_to_source([(1, "a", "None"), (2, "b", "None")])

        o = Orchestrator()
        o.extract_from(
            StreamExtractor(FileHandle.from_tc("AvroSource"), dataset_key="AvroSource")
        )

        o.load_into(
            StreamLoader(
                loader=SimpleLoader(dh_sink, mode="append"),
                options_dict={},
                await_termination=True,
                checkpoint_path=tc.get("AvroSink", "schema_location"),
            )
        )
        o.execute()

        result = DeltaHandle.from_tc("AvroSink").read()

        self.assertEqual(2, result.count())

        # Run again. Should not append more.
        o.execute()
        self.assertEqual(2, result.count())

        self._add_avro_data_to_source([(3, "c", "None"), (4, "d", "None")])

        # Run again. Should append.
        o.execute()
        self.assertEqual(4, result.count())

        # Add specific data to source
        self._add_specific_data_to_source()
        o.execute()
        self.assertEqual(5, result.count())

        # If the same file is altered
        # the new row is appended also
        self._alter_specific_data()
        o.execute()
        self.assertEqual(6, result.count())

    def _create_tbl_mirror(self):
        dh = DeltaHandle.from_tc("MyTblMirror")
        Spark.get().sql(
            f"""
                            CREATE TABLE {dh.get_tablename()}
                            (
                            id int,
                            name string,
                            _rescued_data string
                            )
                            LOCATION '{Configurator().get("MyTblMirror","path")}'
                        """
        )

    def _add_avro_data_to_source(self, input_data: List[Tuple[int, str, str]]):
        df = Spark.get().createDataFrame(
            input_data, "id int, name string, _rescued_data string"
        )

        path = self.avro_source_path + "/" + str(uuid.uuid4())
        df.write.format("avro").save(path)

        for _ in range(20):
            # ensure files are created.
            df = Spark.get().read.format("avro").load(path)
            if df.count() >= len(input_data):
                break
                # all data is in target
            # else continue
            time.sleep(1)
        else:
            # if the loop continues to the end, something went wrong.
            raise Exception(f"Timeout waiting for avro test data in {path}")

    def _add_specific_data_to_source(self):
        df = Spark.get().createDataFrame(
            [(10, "specific", "None")], "id int, name string, _rescued_data string"
        )

        df.write.format("avro").save(self.avro_source_path + "/specific")

    def _alter_specific_data(self):
        df = Spark.get().createDataFrame(
            [(11, "specific", "None")], "id int, name string, _rescued_data string"
        )

        df.write.format("avro").mode("overwrite").save(
            self.avro_source_path + "/specific"
        )



from spetlr.configurator.configurator import Configurator
from spetlr.exceptions import SpetlrException
from spetlr.spark import Spark


class DbHandleException(SpetlrException):
    pass


class DbHandleInvalidName(DbHandleException):
    pass


class DbHandleInvalidFormat(DbHandleException):
    pass


class DbHandle:
    def __init__(self, name: str, location: str = None, data_format: str = "db"):
        self._name = name
        self._location = location
        self._data_format = data_format

        self._validate()

    @classmethod
    def from_tc(cls, id: str):
        tc = Configurator()
        return cls(
            name=tc.table_name(id),
            location=tc.table_property(id, "path", ""),
            data_format=tc.table_property(id, "format", "db"),
        )

    def _validate(self):
        # name is either `db`.`table` or just `table`
        if "." in self._name:
            raise DbHandleInvalidName(f"Invalid DB name {self._name}")

        # only format db is supported.
        if self._data_format != "db":
            raise DbHandleInvalidFormat("Format must be db or null.")

    def drop(self) -> None:
        Spark.get().sql(f"DROP DATABASE IF EXISTS {self._name};")

    def drop_cascade(self) -> None:
        Spark.get().sql(f"DROP DATABASE IF EXISTS {self._name} CASCADE;")

    def create(self) -> None:
        sql = f"CREATE DATABASE IF NOT EXISTS {self._name} "
        if self._location:
            sql += f" LOCATION '{self._location}'"
        Spark.get().sql(sql)



import unittest

from spetlr.utils.GetMergeStatement import GetMergeStatement


class GetMergeStatementTest(unittest.TestCase):
    def test_get_merge_statement_only_with_insert_cols(self):
        output = GetMergeStatement(
            merge_statement_type="delta",
            target_table_name="targetname",
            source_table_name="sourcename",
            join_cols=["col1", "col2"],
            insert_cols=["col1", "col2", "col3", "col4"],
        )

        expected = (
            "MERGE INTO targetname AS target USING sourcename AS source "
            "ON (source.col1 = target.col1) AND (source.col2 = target.col2) "
            "WHEN NOT MATCHED THEN "
            "INSERT (col1, col2, col3, col4) "
            "VALUES (source.col1, source.col2, source.col3, source.col4);"
        )

        self.assertEqual(output, expected)

    def test_get_merge_statement_only_with_update_cols(self):
        output = GetMergeStatement(
            merge_statement_type="delta",
            target_table_name="targetname",
            source_table_name="sourcename",
            join_cols=["col1", "col2"],
            update_cols=["col3", "col4"],
        )

        expected = (
            "MERGE INTO targetname AS target USING sourcename AS source "
            "ON (source.col1 = target.col1) AND (source.col2 = target.col2) "
            "WHEN MATCHED THEN UPDATE "
            "SET target.col3 = source.col3, target.col4 = source.col4;"
        )

        self.assertEqual(output, expected)

    def test_get_merge_statement_with_special_update_set(self):
        output = GetMergeStatement(
            merge_statement_type="delta",
            target_table_name="targetname",
            source_table_name="sourcename",
            join_cols=["col1", "col2"],
            update_cols=["col3", "col4"],
            special_update_set=", target.col5 = test",
        )

        expected = (
            "MERGE INTO targetname AS target USING sourcename AS source "
            "ON (source.col1 = target.col1) AND (source.col2 = target.col2) "
            "WHEN MATCHED THEN UPDATE "
            "SET target.col3 = source.col3, target.col4 = source.col4, "
            "target.col5 = test;"
        )

        self.assertEqual(output, expected)

    def test_get_delta_merge_statement(self):
        output = GetMergeStatement(
            merge_statement_type="delta",
            target_table_name="targetname",
            source_table_name="sourcename",
            join_cols=["col1", "col2"],
            insert_cols=["col1", "col2", "col3", "col4"],
            update_cols=["col3", "col4"],
        )

        expected = (
            "MERGE INTO targetname AS target USING sourcename AS source "
            "ON (source.col1 = target.col1) AND (source.col2 = target.col2) "
            "WHEN MATCHED THEN UPDATE "
            "SET target.col3 = source.col3, target.col4 = source.col4 "
            "WHEN NOT MATCHED THEN "
            "INSERT (col1, col2, col3, col4) "
            "VALUES (source.col1, source.col2, source.col3, source.col4);"
        )

        self.assertEqual(output, expected)

    def test_get_sql_merge_statement(self):
        output = GetMergeStatement(
            merge_statement_type="sql",
            target_table_name="targetname",
            source_table_name="sourcename",
            join_cols=["col1", "col2"],
            insert_cols=["col1", "col2", "col3", "col4"],
            update_cols=["col3", "col4"],
        )

        expected = (
            "MERGE targetname AS target USING sourcename AS source "
            "ON (source.col1 = target.col1) AND (source.col2 = target.col2) "
            "WHEN MATCHED THEN UPDATE "
            "SET target.col3 = source.col3, target.col4 = source.col4 "
            "WHEN NOT MATCHED THEN "
            "INSERT (col1, col2, col3, col4) "
            "VALUES (source.col1, source.col2, source.col3, source.col4);"
        )

        self.assertEqual(output, expected)


if __name__ == "__main__":
    unittest.main()



from spetlr.config_master import Configurator
from spetlr.exceptions import OnlyUseInSpetlrDebugMode
from spetlr.spark import Spark


def CleanupTestDatabases():
    """
    This function can be applied for removing test databases
    in the Databricks Environment.
    """

    c = Configurator()
    if not c.is_debug():
        raise OnlyUseInSpetlrDebugMode()

    id_extension = c.get("ID")

    dbs = Spark.get().sql("SHOW DATABASES").collect()
    for (db,) in dbs:
        if db.endswith(id_extension):
            print(f"Now deleting database {db}")
            Spark.get().sql(f"DROP DATABASE {db} CASCADE")
    print("Database cleanup done.")



from datetime import datetime
from typing import List
from unittest.mock import create_autospec

from spetlrtools.testing import DataframeTestCase

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.etl.loaders.DeleteDataLoader import DeleteDataLoader
from spetlr.functions import get_unique_tempview_name
from spetlr.sql import SqlHandle
from spetlr.sql.SqlServer import SqlServer
from spetlr.utils import DataframeCreator
from tests.cluster.delta import extras
from tests.cluster.delta.SparkExecutor import SparkSqlExecutor


class DeleteDataLoaderDeltaTests(DataframeTestCase):
    target_id = "DeleteDataLoaderDummy"

    input_data = [
        (43, 43.0, "43", datetime(2023, 4, 14, 0, 0, 0)),
        (42, 42.0, "42", datetime(2023, 4, 13, 0, 0, 0)),
        (41, 41.0, "41", datetime(2023, 4, 12, 0, 0, 0)),
        (40, 40.0, "40", datetime(2023, 4, 11, 0, 0, 0)),
    ]

    data1 = [
        (43, 43.0, "43", datetime(2023, 4, 14, 0, 0, 0)),
        (42, 42.0, "42", datetime(2023, 4, 13, 0, 0, 0)),
        (41, 41.0, "41", datetime(2023, 4, 12, 0, 0, 0)),
    ]

    data2 = [
        (42, 42.0, "42", datetime(2023, 4, 13, 0, 0, 0)),
        (41, 41.0, "41", datetime(2023, 4, 12, 0, 0, 0)),
    ]

    data3 = [
        (41, 41.0, "41", datetime(2023, 4, 12, 0, 0, 0)),
    ]

    data4 = []

    dummy_columns: List[str] = ["col1", "col2", "col3", "col4"]

    dummy_schema = None
    target_dh_dummy: DeltaHandle = None

    sql_server_mock = None
    sql_table_handle: SqlHandle = None
    test_table_name = get_unique_tempview_name()

    @classmethod
    def setUpClass(cls) -> None:
        Configurator().add_resource_path(extras)
        Configurator().set_debug()

        cls.target_dh_dummy = DeltaHandle.from_tc("DeleteDataLoaderDummy")

        cls.sql_server_mock = create_autospec(SqlServer)
        cls.sql_table_handle = SqlHandle(
            name=cls.test_table_name, sql_server=cls.sql_server_mock
        )

        SparkSqlExecutor().execute_sql_file("deletedataloader-test")

        cls.dummy_schema = cls.target_dh_dummy.read().schema

        # make sure target is empty
        df_empty = DataframeCreator.make_partial(cls.dummy_schema, [], [])
        cls.target_dh_dummy.overwrite(df_empty)

        # add specific data
        cls.target_dh_dummy.write_or_append(
            df=DataframeCreator.make_partial(
                cls.dummy_schema, cls.dummy_columns, cls.input_data
            ),
            mode="overwrite",
        )

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle.from_tc("DeleteDataLoaderDb").drop_cascade()

    def test_01_can_delete_int_lt(self):
        loader = DeleteDataLoader(
            handle=self.target_dh_dummy,
            comparison_col="col1",
            comparison_limit=41,
            comparison_operator="<",
        )

        loader.save(None)
        self.assertDataframeMatches(self.target_dh_dummy.read(), None, self.data1)

    def test_02_can_delete_float_gt(self):
        loader = DeleteDataLoader(
            handle=self.target_dh_dummy,
            comparison_col="col2",
            comparison_limit=42.0,
            comparison_operator=">",
        )

        loader.save(None)
        self.assertDataframeMatches(self.target_dh_dummy.read(), None, self.data2)

    def test_03_can_delete_string_eq(self):
        loader = DeleteDataLoader(
            handle=self.target_dh_dummy,
            comparison_col="col3",
            comparison_limit="42",
            comparison_operator="=",
        )

        loader.save(None)
        self.assertDataframeMatches(self.target_dh_dummy.read(), None, self.data3)

    def test_04_can_delete_timestamp(self):
        loader = DeleteDataLoader(
            handle=self.target_dh_dummy,
            comparison_col="col4",
            comparison_limit=datetime(2023, 4, 13, 0, 0, 0),
            comparison_operator="<",
        )

        loader.save(None)
        self.assertDataframeMatches(self.target_dh_dummy.read(), None, self.data4)

    def test_05_can_use_sqlhandle(self):
        loader = DeleteDataLoader(
            handle=self.sql_table_handle,
            comparison_col="col1",
            comparison_limit=41,
            comparison_operator="<",
        )

        loader.save(None)

        self.assertEqual(self.sql_server_mock.execute_sql.call_count, 1)



import unittest

from pyspark.sql.types import (
    ArrayType,
    DecimalType,
    FloatType,
    IntegerType,
    LongType,
    MapType,
    StringType,
    StructField,
    StructType,
)

from spetlr.sqlrepr.sql_types import repr_sql_types


class TestReprSqlTypes(unittest.TestCase):
    def test_Primitive(self):
        self.assertEqual(repr_sql_types(StringType()), "StringType()")
        self.assertEqual(repr_sql_types(IntegerType()), "IntegerType()")
        self.assertEqual(repr_sql_types(FloatType()), "FloatType()")
        self.assertEqual(repr_sql_types(LongType()), "LongType()")

    def test_Decimal(self):
        self.assertEqual(
            repr_sql_types(DecimalType(precision=10, scale=2)),
            "DecimalType(precision=10, scale=2)",
        )

    def test_Array(self):
        self.assertEqual(
            repr_sql_types(ArrayType(elementType=IntegerType())),
            "ArrayType(elementType=IntegerType())",
        )

    def test_Map(self):
        self.assertEqual(
            repr_sql_types(MapType(keyType=IntegerType(), valueType=StringType())),
            "MapType(keyType=IntegerType(), "
            "valueType=StringType(), "
            "valueContainsNull=True)",
        )

    def test_StructField(self):
        self.assertEqual(
            repr_sql_types(
                StructField(
                    name="myfield",
                    dataType=IntegerType(),
                    nullable=False,
                    metadata={"foo": "bar"},
                )
            ),
            (
                "StructField(name='myfield', "
                "dataType=IntegerType(), "
                "nullable=False, "
                "metadata={'foo': 'bar'})"
            ),
        )

    def test_Struct(self):
        s = StructType(
            fields=[
                StructField(
                    name="myfield",
                    dataType=IntegerType(),
                    nullable=False,
                ),
                StructField(
                    name="b",
                    dataType=StringType(),
                    nullable=False,
                ),
            ]
        )
        self.assertEqual(
            repr_sql_types(s),
            (
                "StructType(fields=["
                "StructField(name='myfield', dataType=IntegerType(), "
                "nullable=False), "
                "StructField(name='b', dataType=StringType(), "
                "nullable=False)])"
            ),
        )



import unittest

from spetlr.sql import SqlExecutor
from tests.local.sql import sql


class TestSqlExecutor(unittest.TestCase):
    def test_01_standard_splitting(self):
        s = SqlExecutor(sql)
        statements = list(s.get_statements("*"))
        self.assertEqual(len(statements), 3)

    def test_02_marker_splitting(self):
        s = SqlExecutor(sql, statement_spliter=["-- COMMAND ----------"])
        statements = list(s.get_statements("*"))
        self.assertEqual(len(statements), 2)

    def test_03_no_splitting(self):
        s = SqlExecutor(sql, statement_spliter=None)
        statements = list(s.get_statements("*"))
        self.assertEqual(len(statements), 1)

    def test_04_file_by_name(self):
        s = SqlExecutor(sql, statement_spliter=None)
        statements = list(s.get_statements("test1"))
        self.assertEqual(len(statements), 1)

    def test_05_file_not_found(self):
        with self.assertRaises(ValueError):
            s = SqlExecutor(sql, statement_spliter=None)
            list(s.get_statements("unknown"))

    def test_06_file_not_found_and_ignore_empty_folder(self):
        s = SqlExecutor(sql, statement_spliter=None, ignore_empty_folder=True)
        statements = list(s.get_statements("unknown"))
        self.assertEqual(len(statements), 0)



from typing import Union

from pyspark.sql.types import (
    ArrayType,
    DataType,
    DecimalType,
    MapType,
    StructField,
    StructType,
)


def repr_sql_types(obj: Union[StructField, DataType]):
    if isinstance(obj, StructField):
        return (
            (
                f"{obj.__class__.__name__}(name={repr(obj.name)}, "
                f"dataType={repr_sql_types(obj.dataType)}"
            )
            + (f", nullable={repr(obj.nullable)}" if not obj.nullable else "")
            + (f", metadata={repr(obj.metadata)}" if obj.metadata else "")
            + ")"
        )

    if isinstance(obj, ArrayType):
        return (
            f"{obj.__class__.__name__}("
            f"elementType={repr_sql_types(obj.elementType)})"
        )

    if isinstance(obj, StructType):
        fields_part = ", ".join(repr_sql_types(f) for f in obj.fields)
        return f"{obj.__class__.__name__}(" f"fields=[{fields_part}])"

    if isinstance(obj, MapType):
        return (
            f"{obj.__class__.__name__}("
            f"keyType={repr_sql_types(obj.keyType)}, "
            f"valueType={repr_sql_types(obj.valueType)}, "
            f"valueContainsNull={repr(obj.valueContainsNull)})"
        )

    if isinstance(obj, DecimalType):
        DecimalType()
        return (
            f"{obj.__class__.__name__}("
            f"precision={repr(obj.precision)}, "
            f"scale={repr(obj.scale)})"
        )

    # still here? Boring type
    return f"{obj.__class__.__name__}()"



import time
from datetime import datetime, timedelta

import msal
import pandas as pd
import requests
from dateutil import parser
from pytz import timezone, utc

from spetlr.exceptions import SpetlrException

from .PowerBiClient import PowerBiClient


class PowerBi:
    def __init__(
        self,
        client: PowerBiClient,
        *,
        workspace_id: str = None,
        workspace_name: str = None,
        dataset_id: str = None,
        dataset_name: str = None,
        max_minutes_after_last_refresh: int = 12 * 60,
        timeout_in_seconds: int = 15 * 60,
        number_of_retries: int = 0,
        local_timezone_name: str = "UTC",
        ignore_errors: bool = False,
    ):
        """
        Allows refreshing PowerBI datasets and checking if the last refresh
            completed successfully.
        If no workspace is specified, a list of available workspaces will be displayed.
        If no dataset is specified, a list of available datasets in the given workspace
            will be displayed.

        :param PowerBiClient client: PowerBI client credentials.

        :param str workspace_id: The GUID of the workspace.
        :param str workspace_name: The name of the workspace
            (specified instead of the workspace_id).
        :param str dataset_id: The GUID of the dataset.
        :param str dataset_name: The name of the dataset
            (specified instead of the dataset_id).
        :param int max_minutes_after_last_refresh: The number of minutes
            for which the last succeeded refresh is considered valid,
            or 0 to disable time checking. Default is 12 hours.
        :param bool timeout_in_seconds: The number of seconds after which
            the refresh() method times out. Default is 15 minutes.
        :param int number_of_retries: The number of retries on transient
            errors when calling refresh(). Default is 0 (no retries).
            (e.g. 1 means two attempts in total.)
            Used only when the timeout_in_seconds parameter allows it!
        :param str local_timezone_name: The timezone to use when showing
            refresh timestamps. Only used for printing timestamps.
            Default timezone is UTC.
        :param bool ignore_errors: True to print errors in the output
            or False (default) to cast a SpetlrException.
        """

        if workspace_id is not None and workspace_name is not None:
            raise ValueError("Specify either 'workspace_id' or 'workspace_name'!")
        if dataset_id is not None and dataset_name is not None:
            raise ValueError("Specify either 'dataset_id' or 'dataset_name'!")
        if max_minutes_after_last_refresh is None or max_minutes_after_last_refresh < 0:
            raise ValueError(
                "The 'max_minutes_after_last_refresh' parameter "
                "must be greater than or equal zero!"
            )
        if timeout_in_seconds is None or timeout_in_seconds <= 0:
            raise ValueError(
                "The 'timeout_in_seconds' parameter must be greater than zero!"
            )
        if number_of_retries is None or number_of_retries < 0:
            raise ValueError(
                "The 'number_of_retries' parameter "
                "must be greater than or equal zero!"
            )

        self.workspace_id = workspace_id
        self.workspace_name = workspace_name
        self.dataset_id = dataset_id
        self.dataset_name = dataset_name

        # Set access parameters
        self.client = client

        self.max_minutes_after_last_refresh = max_minutes_after_last_refresh
        self.timeout_in_seconds = timeout_in_seconds
        self.number_of_retries = number_of_retries
        self.local_timezone_name = (
            local_timezone_name if local_timezone_name is not None else "UTC"
        )
        self.ignore_errors = ignore_errors
        self.api_header = None
        self.expire_time = 0

        self.last_status = None
        self.last_exception = None
        self.last_refresh_utc = None
        self.last_duration_in_seconds = 0

    def _raise_error(self, message: str) -> None:
        if self.ignore_errors:
            print(message)
        else:
            raise SpetlrException(message)

    def _raise_api_error(self, message: str, api_call: requests.Response) -> None:
        print(api_call.text)
        self._raise_error(
            message + f" Response: {api_call.status_code} {api_call.reason}"
        )

    def _get_access_token(self) -> bool:
        """
        Acquires an access token to connect to PowerBI.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        if self.expire_time != 0:
            if time.time() < self.expire_time:
                return True
            print("Renewing access token...")

        # Prepare URLs
        authority_url = f"https://login.microsoftonline.com/{self.client.tenant_id}/"
        scope = ["https://analysis.windows.net/powerbi/api/.default"]
        self.powerbi_url = "https://api.powerbi.com/v1.0/myorg/"

        # Use MSAL to get an access token
        app = msal.ConfidentialClientApplication(
            self.client.client_id,
            authority=authority_url,
            client_credential=self.client.client_secret,
        )
        result = app.acquire_token_for_client(scopes=scope)
        if "access_token" not in result:
            print(result)
            self._raise_error("Failed to acquire token!")
            return False

        default_expires_in = 5 * 60
        extra_seconds = 3

        access_token = result["access_token"]
        if "expires_in" in result:
            self.expire_time = time.time() + result["expires_in"] - extra_seconds
        else:
            self.expire_time = time.time() + default_expires_in - extra_seconds

        self.api_header = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {access_token}",
        }
        return True

    @staticmethod
    def _show_workspaces(df: pd.DataFrame) -> None:
        print("Available workspaces:")
        df.rename(
            columns={"id": "workspace_id", "name": "workspace_name"},
            inplace=True,
        )
        df.display()

    @staticmethod
    def _show_datasets(df: pd.DataFrame) -> None:
        print("Available datasets:")
        df.rename(
            columns={"id": "dataset_id", "name": "dataset_name"},
            inplace=True,
        )
        df.display()

    def _get_workspace(self) -> bool:
        """
        Gets workspace ID based on the workspace name, or shows all workspaces
            if no parameter was specified.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        if self.workspace_id is None:
            api_call = requests.get(
                url=f"{self.powerbi_url}groups", headers=self.api_header
            )
            if api_call.status_code != 200:
                self._raise_api_error("Failed to fetch workspaces!", api_call)
                return False
            df = pd.DataFrame(api_call.json()["value"], columns=["id", "name"])

            if self.workspace_name is None:
                self._show_workspaces(df)
                return False

            rows = df.loc[df["name"] == self.workspace_name, "id"]
            if rows.empty:
                self._raise_error(
                    f"Workspace name '{self.workspace_name}' cannot be found!"
                )
                return False
            self.workspace_id = rows.values[0]

        return True

    def _get_dataset(self) -> bool:
        """
        Gets dataset ID based on the dataset name, or shows all datasets
            if no parameter was specified.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        if self.dataset_id is None:
            api_call = requests.get(
                url=f"{self.powerbi_url}groups/{self.workspace_id}/datasets",
                headers=self.api_header,
            )
            if api_call.status_code != 200:
                self._raise_api_error("Failed to fetch datasets!", api_call)
                return False
            df = pd.DataFrame(api_call.json()["value"], columns=["id", "name"])

            if self.dataset_name is None:
                self._show_datasets(df)
                return False

            rows = df.loc[df["name"] == self.dataset_name, "id"]
            if rows.empty:
                self._raise_error(
                    f"Dataset name '{self.dataset_name}' cannot be found, "
                    "or the dataset doesn't have a user with the required permissions!"
                )
                return False
            self.dataset_id = rows.values[0]

        return True

    def _connect(self) -> bool:
        """
        Connects or reconnects to PowerBI to fetch refresh history or trigger a refresh.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        if not self._get_access_token():
            return False
        if not self._get_workspace():
            return False
        if not self._get_dataset():
            return False

        return True

    def _get_last_refresh(self) -> bool:
        """
        Gets the latest record in the PowerBI dataset refresh history.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        if not self._connect():
            return False

        self.last_status = None
        self.last_exception = None
        self.last_refresh_utc = None
        self.last_duration_in_seconds = 0

        # Note: we fetch only the latest refresh record, i.e. top=1
        api_url = (
            f"{self.powerbi_url}groups/{self.workspace_id}"
            f"/datasets/{self.dataset_id}/refreshes?$top=1"
        )
        api_call = requests.get(url=api_url, headers=self.api_header)
        if api_call.status_code == 200:
            json = api_call.json()
            df = pd.DataFrame(
                json["value"],
                columns=[
                    "requestId",
                    "id",
                    "refreshType",
                    "startTime",
                    "endTime",
                    "status",
                    "serviceExceptionJson",
                ],
            )
            if not df.empty:
                df.set_index("id")
                self.last_status = df.status[0]
                self.last_exception = df.serviceExceptionJson[0]
                start_time = df.startTime[0]
                end_time = df.endTime[0]
                if (
                    self.last_status == "Completed"
                    and end_time is not None
                    and len(end_time) > 0
                ):
                    self.last_refresh_utc = parser.parse(end_time).replace(tzinfo=utc)
                    if start_time is not None and len(start_time) > 0:
                        self.last_duration_in_seconds = int(
                            (
                                parser.parse(end_time) - parser.parse(start_time)
                            ).total_seconds()
                        )
            return True
        elif api_call.status_code == 404:
            self._raise_error(
                "The specified dataset or workspace cannot be found, "
                "or the dataset doesn't have a user with the required permissions!"
            )
        else:
            self._raise_api_error("Failed to fetch refresh history!", api_call)
        return False

    def _verify_last_refresh(self) -> bool:
        """
        Checks if the last refresh of the PowerBI dataset completed successfully,
            and verifies if it happened recently enough.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        if self.last_status is None:
            self._raise_error("Refresh is still in progress or never triggered!")
        elif self.last_status == "Completed":
            if self.last_refresh_utc is None:
                self._raise_error("Completed at unknown refresh time!")
            else:
                last_refresh_str = self.last_refresh_utc.astimezone(
                    timezone(self.local_timezone_name)
                ).strftime("%Y-%m-%d %H:%M") + (
                    " (UTC)"
                    if self.local_timezone_name.lower() == "utc"
                    else " (local time)"
                )
                min_refresh_time_utc = datetime.now(utc) - timedelta(
                    minutes=self.max_minutes_after_last_refresh
                )
                if (
                    self.max_minutes_after_last_refresh > 0
                    and self.last_refresh_utc < min_refresh_time_utc
                ):
                    self._raise_error(
                        "Last refresh finished more than "
                        f"{self.max_minutes_after_last_refresh} "
                        f"minutes ago at {last_refresh_str} !"
                    )
                else:
                    print(f"Refresh completed successfully at {last_refresh_str}.")
                    return True
        elif self.last_status == "Unknown":
            self._raise_error("Refresh is still in progress!")
        elif self.last_status == "Disabled":
            self._raise_error("Refresh is disabled!")
        elif self.last_status == "Failed":
            self._raise_error(f"Last refresh failed! {self.last_exception}")
        else:
            self._raise_error(
                f"Unknown refresh status: {self.last_status}! {self.last_exception}"
            )
        return False

    def _trigger_new_refresh(self) -> bool:
        """
        Starts a refresh of the PowerBI dataset.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        if self.last_status is None or self.last_status in ["Completed", "Failed"]:
            if self.last_status == "Failed":
                print(f"Warning: Last refresh failed! {self.last_exception}")
                print()

            api_url = (
                f"{self.powerbi_url}groups/{self.workspace_id}"
                f"/datasets/{self.dataset_id}/refreshes"
            )
            api_call = requests.post(url=api_url, headers=self.api_header)
            if api_call.status_code == 202:
                print("A new refresh has been successfully triggered.")
                return True
            else:
                self._raise_api_error("Failed to trigger a refresh!", api_call)
        elif self.last_status == "Unknown":
            print("Refresh is already in progress!")
            return True
        elif self.last_status == "Disabled":
            self._raise_error("Refresh is disabled!")
        else:
            self._raise_error(
                f"Unknown refresh status: {self.last_status}! {self.last_exception}"
            )
        return False

    def _get_seconds_to_wait(self, elapsed_seconds: int) -> int:
        """
        Returns the number of seconds to wait before rechecking
            if the refresh has completed. The method makes sure as few requests
            to the PowerBI API as possible would be made.

        :return: number of seconds to wait
        :rtype: int
        """

        # Microsoft set a limit of how many API requests in an hour can be made
        # to the PowerBI API. This logic is a fine compromise between how often
        # (small delay) we check through polling if the refresh has finished and
        # how soon (large delay) will the refresh() method complete.

        remaining_seconds = self.timeout_in_seconds - elapsed_seconds

        wait_seconds = elapsed_seconds
        if self.last_duration_in_seconds > 0:
            wait_seconds = abs(self.last_duration_in_seconds - elapsed_seconds)

        wait_seconds = (
            15 if wait_seconds < 40 else 60 if wait_seconds < (3 * 60) else (5 * 60)
        )

        return (
            remaining_seconds
            if (0 < remaining_seconds < wait_seconds)
            else wait_seconds
        )

    def check(self) -> bool:
        """
        Checks if the last refresh of the PowerBI dataset completed successfully,
            and verifies if it happened recently enough.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        return self._get_last_refresh() and self._verify_last_refresh()

    def start_refresh(self) -> bool:
        """
        Starts a refresh of the PowerBI dataset without waiting.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        return self._get_last_refresh() and self._trigger_new_refresh()

    def refresh(self) -> bool:
        """
        Starts a refresh of the PowerBI dataset and waits until completed.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises SpetlrException: if failed and ignore_errors==False
        """

        retries = self.number_of_retries
        start_time = time.time()
        if not self.start_refresh():
            return False

        while True:
            elapsed_seconds = int(time.time() - start_time)
            if elapsed_seconds >= self.timeout_in_seconds:
                print("Timeout!")
                break
            wait_seconds = self._get_seconds_to_wait(elapsed_seconds)
            print(f"Waiting {wait_seconds} seconds...")
            time.sleep(wait_seconds)

            if not self._get_last_refresh():
                return False

            if self.last_status == "Failed" and retries > 0:
                retries = retries - 1
                if not self._trigger_new_refresh():
                    return False
                continue

            if self.last_status != "Unknown":
                break

        return self._verify_last_refresh()



import unittest

from spetlrtools.testing import TestHandle

from spetlr.etl import Orchestrator
from spetlr.etl.extractors import LazySimpleExtractor


class TestLazyExtractor(unittest.TestCase):
    def test_lazy_extractor(self):
        # This is the orchestrator. It takes any lunch it is given
        # and returns it. If none is given, it extracts bacon and
        # returns it.
        o = Orchestrator()
        o.extract_from(LazySimpleExtractor(TestHandle(provides="bacon"), "lunch"))

        # Given spam? return spam
        lazy_result = o.execute({"lunch": "spam"})
        self.assertEqual(lazy_result["lunch"], "spam")

        # no lunch? use bacon
        active_result = o.execute()
        self.assertEqual(active_result["lunch"], "bacon")



import unittest
from unittest.mock import MagicMock

from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from spetlr.etl import Loader
from spetlr.spark import Spark


class LoaderTests(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.loader = Loader()
        self.loader.save = MagicMock()
        self.loader.save_many = MagicMock()

        self.df_1 = create_dataframe()
        self.df_2 = create_dataframe()
        self.df_3 = create_dataframe()

    def test_return_inputs(self):
        # Assert Loader returns ouput
        input = {"my_key_1": self.df_1}
        result = self.loader.etl(input)
        self.assertIs(input, result)

    def test_save(self):
        # save is called when:
        # - inputs has one df
        # - dataset_input_keys is not given
        self.loader.etl({"my_key_1": self.df_1})

        self.assertTrue(self.loader.save.called)
        self.assertFalse(self.loader.save_many.called)

        # the .args part became available in python 3.8
        args = self.loader.save.call_args[0]
        self.assertEqual(len(args), 1)
        self.assertIs(args[0], self.df_1)

    def test_save_many(self):
        # save is called when:
        # - inputs has more than one df
        # - dataset_input_keys is given
        self.loader.etl({"my_key_1": self.df_1, "my_key_2": self.df_2})

        self.assertFalse(self.loader.save.called)
        self.assertTrue(self.loader.save_many.called)

        # the .args part became available in python 3.8
        args = self.loader.save_many.call_args[0]
        self.assertEqual(len(args), 1)

        datasets = args[0]
        self.assertEqual(len(datasets), 2)
        self.assertEqual(set(datasets.keys()), {"my_key_1", "my_key_2"})
        self.assertEqual(set(datasets.values()), {self.df_1, self.df_2})

    def test_save_input_key(self):
        # save is called when:
        # - dataset_input_keys is str

        loader = Loader(dataset_input_keys="my_key_1")
        loader.save = MagicMock()
        loader.save_many = MagicMock()

        loader.etl({"my_key_1": self.df_1, "my_key_2": self.df_2})

        self.assertTrue(loader.save.called)
        self.assertFalse(loader.save_many.called)

        # the .args part became available in python 3.8
        args = loader.save.call_args[0]
        self.assertEqual(len(args), 1)
        self.assertIs(args[0], self.df_1)

    def test_save_single_input_key_list(self):
        # save is called when:
        # - dataset_input_keys has len 1

        loader = Loader(dataset_input_keys=["my_key_1"])
        loader.save = MagicMock()
        loader.save_many = MagicMock()

        loader.etl({"my_key_1": self.df_1, "my_key_2": self.df_2})

        self.assertTrue(loader.save.called)
        self.assertFalse(loader.save_many.called)

        # the .args part became available in python 3.8
        args = loader.save.call_args[0]
        self.assertEqual(len(args), 1)
        self.assertIs(args[0], self.df_1)

    def test_save_many_input_key_list(self):
        # save_many is called when:
        # - dataset_input_keys has len more than 1

        loader = Loader(dataset_input_keys=["my_key_1", "my_key_2"])
        loader.save = MagicMock()
        loader.save_many = MagicMock()

        loader.etl(
            {"my_key_1": self.df_1, "my_key_2": self.df_2, "my_key_3": self.df_3}
        )

        self.assertFalse(loader.save.called)
        self.assertTrue(loader.save_many.called)

        # the .args part became available in python 3.8
        args = loader.save_many.call_args[0]
        self.assertEqual(len(args), 1)

        datasets = args[0]
        self.assertEqual(len(datasets), 2)
        self.assertEqual(set(datasets.keys()), {"my_key_1", "my_key_2"})
        self.assertEqual(set(datasets.values()), {self.df_1, self.df_2})


def create_dataframe():
    return Spark.get().createDataFrame(
        Spark.get().sparkContext.parallelize([(1, "1"), (2, "2"), (3, "3")]),
        StructType(
            [
                StructField("id", IntegerType(), False),
                StructField("text", StringType(), False),
            ]
        ),
    )


if __name__ == "__main__":
    unittest.main()



from typing import List, Protocol, Union

from pyspark.sql import DataFrame


class Upsertable(Protocol):
    def upsert(self, df: DataFrame, join_cols: List[str]) -> Union[DataFrame, None]:
        pass


class Overwritable(Protocol):
    def overwrite(
        self, df: DataFrame, mergeSchema: bool = None, overwriteSchema: bool = None
    ) -> None:
        pass


class Appendable(Protocol):
    def append(self, df: DataFrame, mergeSchema: bool = None) -> None:
        pass



from spetlr.exceptions import SpetlrException


class SpetlrEhException(SpetlrException):
    pass


class SpetlrEhInitException(SpetlrEhException):
    pass


class SpetlrEhLogicException(SpetlrEhException):
    pass



from pyspark.sql import DataFrame

from spetlr.etl import Extractor
from spetlr.exceptions import SchemasNotEqualException
from spetlr.tables.TableHandle import TableHandle


class CheckSchemaExtractor(Extractor):
    """This extractor will read from any TableHandle and
    test the read dataframe against a schema stored in the object.
    The schemas are compared for changes.
    """

    def __init__(self, handle: TableHandle, dataset_key: str = None):
        super().__init__(dataset_key=dataset_key)
        self.handle = handle

    def read(self) -> DataFrame:
        # get expected schema from handle
        expected_schema = self.handle.get_schema()
        # remove schema from handle to not use it for reading data
        self.handle.set_schema(None)
        # read data without schema from code
        data_schema = self.handle.read().schema
        # set schema back in handler
        self.handle.set_schema(expected_schema)

        if not data_schema.__eq__(expected_schema):
            raise SchemasNotEqualException(
                f"""Schemas have different number of columns.
                Data schema:
                {data_schema.json()}
                Expected schema:
                {expected_schema.json()}
                """
            )



import unittest
from datetime import datetime, timedelta
from unittest.mock import Mock, patch

from pytz import utc

from spetlr.exceptions import SpetlrException
from spetlr.power_bi.PowerBi import PowerBi
from spetlr.power_bi.PowerBiClient import PowerBiClient


class TestPowerBi(unittest.TestCase):
    @patch("requests.get")
    def test_get_workspace_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "value": [{"id": "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0", "name": "Finance"}]
        }
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_name="Finance")
        sut.powerbi_url = "test"

        # Act
        result = sut._get_workspace()

        # Assert
        self.assertTrue(result)
        self.assertEqual(sut.workspace_id, "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0")

    @patch("requests.get")
    def test_get_workspace_failure(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 404
        mock_response.text = "error"
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_name="Finance")
        sut.powerbi_url = "test"

        # Act
        with self.assertRaises(SpetlrException) as context:
            sut._get_workspace()

        # Assert
        self.assertIn("Failed to fetch workspaces", str(context.exception))

    @patch("requests.get")
    def test_get_dataset_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "value": [
                {"id": "b1f0a07e-e348-402c-a2b2-11f3e31181ce", "name": "Invoicing"}
            ]
        }
        mock_get.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_name="Finance",
            dataset_name="Invoicing",
        )
        sut.powerbi_url = "test"

        # Act
        result = sut._get_dataset()

        # Assert
        self.assertTrue(result)
        self.assertEqual(sut.dataset_id, "b1f0a07e-e348-402c-a2b2-11f3e31181ce")

    @patch("requests.get")
    def test_get_dataset_failure(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 404
        mock_response.text = "error"
        mock_get.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_name="Finance",
            dataset_name="Invoicing",
        )
        sut.powerbi_url = "test"

        # Act
        with self.assertRaises(SpetlrException) as context:
            sut._get_dataset()

        # Assert
        self.assertIn("Failed to fetch datasets", str(context.exception))

    @patch("requests.get")
    def test_get_last_refresh_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "value": [
                {
                    "requestId": "1",
                    "id": "1",
                    "refreshType": "Manual",
                    "startTime": "2024-02-26T10:00:00Z",
                    "endTime": "2024-02-26T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                }
            ]
        }
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_id="test", dataset_id="test")
        sut.powerbi_url = "test"
        sut._connect = lambda: True

        # Act
        result = sut._get_last_refresh()

        # Assert
        self.assertTrue(result)
        self.assertEqual(sut.last_status, "Completed")
        self.assertIsNone(sut.last_exception)
        self.assertEqual(str(sut.last_refresh_utc), "2024-02-26 10:05:00+00:00")
        self.assertEqual(sut.last_duration_in_seconds, 5 * 60)

    @patch("requests.get")
    def test_get_last_refresh_failure(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 404  # dataset or workspace not found
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_id="test", dataset_id="test")
        sut.powerbi_url = "test"
        sut._connect = lambda: True

        # Act
        with self.assertRaises(SpetlrException) as context:
            sut._get_last_refresh()

        # Assert
        self.assertIn(
            "The specified dataset or workspace cannot be found",
            str(context.exception),
        )

    def test_verify_last_refresh_success(self):
        # Arrange
        sut = PowerBi(
            PowerBiClient(),
            workspace_id="test",
            dataset_id="test",
            max_minutes_after_last_refresh=5,
            local_timezone_name="Europe/Copenhagen",
        )
        sut.last_status = "Completed"
        sut.last_refresh_utc = datetime.now(utc) - timedelta(minutes=1)

        # Act
        result = sut._verify_last_refresh()

        # Assert
        self.assertTrue(result)

    def test_verify_last_refresh_failure(self):
        # Arrange
        sut = PowerBi(
            PowerBiClient(),
            workspace_id="test",
            dataset_id="test",
            max_minutes_after_last_refresh=5,
            local_timezone_name="Europe/Copenhagen",
        )
        sut.last_status = "Completed"
        sut.last_refresh_utc = datetime.now(utc) - timedelta(minutes=10)

        # Act
        with self.assertRaises(SpetlrException) as context:
            sut._verify_last_refresh()

        # Assert
        self.assertIn("Last refresh finished more than", str(context.exception))

    @patch("requests.post")
    def test_trigger_new_refresh_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 202
        mock_get.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
        )
        sut.last_status = "Completed"
        sut.powerbi_url = "test"

        # Act
        result = sut._trigger_new_refresh()

        # Assert
        self.assertTrue(result)

    @patch("requests.post")
    def test_trigger_new_refresh_failure(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 404
        mock_response.text = "error"
        mock_get.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
        )
        sut.last_status = "Completed"
        sut.powerbi_url = "test"

        # Act
        with self.assertRaises(SpetlrException) as context:
            sut._trigger_new_refresh()

        # Assert
        self.assertIn("Failed to trigger a refresh", str(context.exception))

    def test_get_seconds_to_wait_first(self):
        # Arrange
        sut = PowerBi(PowerBiClient())
        sut.last_duration_in_seconds = 0
        elapsed = 5
        expected_result = 15

        # Act
        result = sut._get_seconds_to_wait(elapsed)

        # Assert
        self.assertEqual(result, expected_result)

    def test_get_seconds_to_wait_next(self):
        # Arrange
        sut = PowerBi(PowerBiClient())
        sut.last_duration_in_seconds = 15 * 60
        elapsed = 5
        expected_result = 5 * 60

        # Act
        result = sut._get_seconds_to_wait(elapsed)

        # Assert
        self.assertEqual(result, expected_result)

    def test_get_seconds_to_wait_exceeding_timeout(self):
        # Arrange
        sut = PowerBi(PowerBiClient(), timeout_in_seconds=90)
        sut.last_duration_in_seconds = 15 * 60
        elapsed = 5
        expected_result = 90 - elapsed

        # Act
        result = sut._get_seconds_to_wait(elapsed)

        # Assert
        self.assertEqual(result, expected_result)



from typing import Any, Dict

import pyspark.sql.types as T
from pyspark.sql import DataFrame

from spetlr.configurator.configurator import Configurator
from spetlr.delta.delta_handle import DeltaHandleInvalidFormat
from spetlr.exceptions import SparkVersionNotSupportedForSpetlrStreaming
from spetlr.schema_manager import SchemaManager
from spetlr.spark import Spark
from spetlr.tables import TableHandle


class FileHandle(TableHandle):
    def __init__(
        self,
        *,
        file_location: str,
        data_format: str,
        options: Dict[str, Any] = None,
        schema_location: str = None,
        schema: T.StructType = None,
    ):
        """
        file_location: the path to the file location of the delta table
        schema_location: The location of the cloudfile schema.
            Databricks documentation uses the expected writeStream checkpoint_path
            as schema_location, see documentation at
            https://docs.databricks.com/getting-started/etl-quick-start.html#auto-loader
        data_format: The expected data format extracted by the Autoloader.
            Examples: JSON, CSV, PARQUET, AVRO, ORC, TEXT, and BINARYFILE.
            See: https://docs.databricks.com/ingestion/auto-loader/index.html
        schema: Pyspark schema to use for data extraction
        """

        if Spark.version() < Spark.DATABRICKS_RUNTIME_10_4:
            raise SparkVersionNotSupportedForSpetlrStreaming()

        self._location = file_location
        self._data_format = data_format

        self._options = options or {}
        self._schema_location = schema_location
        self._schema = schema

        self._options.update(
            {
                "cloudFiles.format": self._data_format,
            }
        )

        if schema_location:
            self._options.update(
                {
                    "cloudFiles.schemaLocation": self._schema_location,
                }
            )

        self._validate()

    @classmethod
    def from_tc(cls, id: str) -> "FileHandle":
        tc = Configurator()
        sm = SchemaManager()
        return cls(
            file_location=tc.get(id, "path"),
            data_format=tc.get(id, "format"),
            schema_location=tc.get(id, "schema_location", None),
            schema=sm.get_schema(id, None),
        )

    def _validate(self):
        """
        Validates the dataformat is not delta,
        since delta stream should use the DeltaHandle instead.
        """
        if self._data_format == "delta":
            raise DeltaHandleInvalidFormat(
                "Use DeltaHandle.read() or DeltaHandle.read_stream() for delta."
            )

    def read(self) -> DataFrame:
        reader = Spark.get().read.format(self._data_format)

        if self._options:
            reader = reader.options(**self._options)

        if self._schema:
            reader = reader.schema(self._schema)

        return reader.load(self._location)

    def read_stream(self) -> DataFrame:
        reader = Spark.get().readStream.format("cloudFiles")

        if self._options:
            reader = reader.options(**self._options)

        if self._schema:
            reader = reader.schema(self._schema)

        return reader.load(self._location)

    def get_schema(self) -> T.StructType:
        return self._schema

    def set_schema(self, schema: T.StructType) -> "FileHandle":
        self._schema = schema

        return self

    def get_options(self) -> Dict[str, Any]:
        return self._options

    def set_options(self, options: Dict[str, Any]) -> "FileHandle":
        self._options.update(options)

        return self



"""
A collection of useful pyspark snippets.
"""

import hashlib
import json
import uuid as _uuid
from typing import Any

from deprecated import deprecated
from pyspark.sql.functions import udf
from pyspark.sql.types import StringType

from spetlr.exceptions import NoDbUtils, NoTableException
from spetlr.spark import Spark

# Pyspark uuid function implemented as recommended here
# https://stackoverflow.com/questions/49785108/spark-streaming-with-python-how-to-add-a-uuid-column/50095755
uuid = udf(lambda: str(_uuid.uuid4()), StringType()).asNondeterministic()  # noqa


def init_dbutils():
    try:
        from pyspark.dbutils import DBUtils

        return DBUtils(Spark.get())
    except (ImportError, ModuleNotFoundError):
        try:
            import IPython

            return IPython.get_ipython().user_ns["dbutils"]
        except (ImportError, ModuleNotFoundError):
            raise NoDbUtils()


@deprecated(reason="Use DeltaHandle.drop_and_delete")
def drop_table_cascade(DBDotTableName: str) -> None:
    """
    drop_table_cascade drops a databricks database table
    and remove the files permanently

    :param DBDotTableName: db.tablename

    return None

    """

    # Check if table exists
    if not Spark.get()._jsparkSession.catalog().tableExists(DBDotTableName):
        raise NoTableException(
            f"The table {DBDotTableName} not found. Remember syntax db.tablename."
        )

    # Get table path
    table_path = str(
        Spark.get().sql(f"DESCRIBE DETAIL {DBDotTableName}").collect()[0]["location"]
    )

    if table_path is None:
        raise NoTableException("Table path is NONE.")

    # Remove files associated with table
    init_dbutils().fs.rm(table_path, recurse=True)

    # Remove table
    Spark.get().sql(f"DROP TABLE IF EXISTS {DBDotTableName}")


def get_unique_tempview_name() -> str:
    unique_id = _uuid.uuid4().hex
    temp_view_name = f"source_{unique_id}"
    return temp_view_name


def json_hash(value: Any):
    """
    Generate a hash from the json representation of a simple type.

    This function converts the input to a JSON string,
    generates a SHA256 hash of that string, and returns the hex digest.
    Note that the function sorts any dictionary keys before converting
    to a string to ensure consistent output for the same dictionary
    regardless of the original key order.

    Args:
        value: A simple type or dictionary of simple types (e.g. int, str).

    Returns:
        str: A string representation of the input object.

    Raises:
        TypeError: If the value cannot be serialized to JSON.
    """
    # Convert the dictionary to a JSON string and encode it to bytes
    input_bytes = json.dumps(value, sort_keys=True).encode("utf-8")

    # Generate a SHA256 hash of the input
    hash_object = hashlib.sha256(input_bytes)

    # Get the hexadecimal representation of the hash
    return hash_object.hexdigest()



class PowerBiClient:
    def __init__(
        self,
        *,
        client_id: str = None,
        client_secret: str = None,
        tenant_id: str = None,
    ):
        """
        Creates PowerBI client credentials.

        :param str client_id: PowerBI client ID.
        :param str client_secret: PowerBI client secret.
        :param str tenant_id: PowerBI tenant ID.
        """

        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id



import re
from typing import Dict

import sqlparse

from spetlr.exceptions.configurator_exceptions import (
    SpetlrConfiguratorInvalidSqlException,
)

from .exceptions import _UnpackAttemptFailed
from .StatementBlocks import ClusteredBy, StatementBlocks
from .substructures import (
    _unpack_list_of_single_variables,
    _unpack_list_of_sorting_statements,
    _unpack_options,
)
from .types import _PeekableTokenList


def _walk_table_statement(statement: _PeekableTokenList) -> Dict:
    # for tables, we need to collect all tokens until the schema starts.
    name_parts = []
    blocks = StatementBlocks()
    while True:
        # until we managed to unpack the rest of the table,
        # all the bits must be part of the name.
        # This allows unpacking names like CREATE TABLE {MYDB}.tbl{ID}
        # which will return lots of parsing errors otherwise.

        name_parts.append(next(statement).value.strip("`"))

        try:
            blocks = _extract_table_blocks(statement)
        except _UnpackAttemptFailed:
            continue
        break

    object_details = blocks.get_simple_structure()
    object_details["name"] = "".join(name_parts)

    return object_details


def _extract_table_blocks(stmt: _PeekableTokenList) -> StatementBlocks:
    """After the initial "create table [if not exists] mytable",
    The statements can contain several blocks that are mostly optional,
    (check the documentation links at the top of this file)
    and which can come in any order.
    (we are not validating SQL here, only extracting information)
    This function extracts those blocks and makes them available.

    (If the unpacking fails before the first token is consumed,
    raise the _UnpackAttemptFailed exception, to allow further
    processing.)
    """

    blocks = StatementBlocks()

    if isinstance(stmt.peek(), sqlparse.sql.Parenthesis):
        blocks.schema = next(stmt).value
    else:
        raise _UnpackAttemptFailed("No valid table schema")

    for token in stmt:
        val = token.value.upper()

        if val == "OPTIONS":
            blocks.options = {}
            for k, v in _unpack_options(stmt):
                blocks.options[k] = v
            continue

        if val == "USING":
            blocks.using = next(stmt).value
            continue

        if re.match(r"PARTITIONED\s+BY", val):
            try:
                blocks.partitioned_by = [
                    token.value for token in _unpack_list_of_single_variables(stmt)
                ]
            except SpetlrConfiguratorInvalidSqlException:
                raise SpetlrConfiguratorInvalidSqlException(
                    "Please add your partition columns to the schema."
                )
            continue

        if re.match(r"CLUSTERED\s+BY", val):
            blocks.clustered_by = ClusteredBy(
                [token.value for token in _unpack_list_of_single_variables(stmt)]
            )

            peek = next(stmt)
            if re.match(r"SORTED\s+BY", peek.value.upper()):
                blocks.clustered_by.sorting = []
                for sort_col in _unpack_list_of_sorting_statements(stmt):
                    blocks.clustered_by.sorting.append(sort_col)
                into_token = next(stmt)
            else:
                into_token = peek

            if not into_token.value.upper() == "INTO":
                SpetlrConfiguratorInvalidSqlException("expected INTO after clustering")

            try:
                blocks.clustered_by.n_buckets = int(next(stmt).value)
            except ValueError:
                raise SpetlrConfiguratorInvalidSqlException(
                    "Unable to extract ordering buckets"
                )

            buckets_token = next(stmt)
            if not buckets_token.value.upper() == "BUCKETS":
                SpetlrConfiguratorInvalidSqlException("expected BUCKETS after INTO n")
            continue

        if val == "LOCATION":
            blocks.location = next(stmt).value.strip("\"'")
            continue

        if val == "COMMENT":
            blocks.comment = next(stmt).value.strip("\"'")
            continue

        if val == "TBLPROPERTIES":
            blocks.tblproperties = {}
            for k, v in _unpack_options(stmt):
                blocks.tblproperties[k] = v
            continue

        if val == "AS":
            raise NotImplementedError("Select AS not currently supported")

        if val == ";":
            break

        # if we got to here with no exception, break or continue something is wrong
        raise SpetlrConfiguratorInvalidSqlException(
            "Unknown statement form encountered."
        )
    return blocks



import unittest

from pyspark.sql.types import IntegerType, StructField, StructType
from spetlrtools.testing import DataframeTestCase, TestHandle

from spetlr.etl.extractors import SimpleExtractor
from spetlr.etl.log import LogOrchestrator
from spetlr.etl.log.log_transformers import CountLogTransformer, NullLogTransformer
from spetlr.utils import DataframeCreator


class TestLogTransformer(DataframeTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        schema = StructType(
            [
                StructField("col_1", IntegerType(), True),
                StructField("col_2", IntegerType(), True),
                StructField("col_3", IntegerType(), True),
            ]
        )

        data = [
            (1, 1, 1),
            (2, 2, 2),
            (3, 3, 3),
            (4, 4, 4),
            (5, 5, 5),
        ]

        cls.df = DataframeCreator.make(schema, data)

    def test_log_step_single_df_01(self) -> None:
        sink_log_handle = TestHandle()
        source_handle = TestHandle(self.df)

        log_oc = LogOrchestrator(
            handles=[sink_log_handle],
        )

        log_oc.extract_from(
            SimpleExtractor(
                handle=source_handle,
                dataset_key="df_test",
            )
        )

        log_oc.log_with(
            CountLogTransformer(
                log_name="log_test",
                dataset_input_keys=["df_test"],
            )
        )

        log_oc.execute()

        df_log = sink_log_handle.appended

        self.assertIsNotNone(df_log)
        self.assertEqual(df_log.count(), 1)

    def test_log_step_multiple_dfs_02(self) -> None:
        sink_log_handle = TestHandle()
        source_handle = TestHandle(self.df)

        log_oc = LogOrchestrator(
            handles=[sink_log_handle],
        )

        log_oc.extract_from(
            SimpleExtractor(
                handle=source_handle,
                dataset_key="df_test_1",
            )
        )

        log_oc.extract_from(
            SimpleExtractor(
                handle=source_handle,
                dataset_key="df_test_2",
            )
        )

        log_oc.log_with(
            CountLogTransformer(
                log_name="log_test_count",
                dataset_input_keys=["df_test_1", "df_test_2"],
                consume_inputs=False,
            )
        )

        log_oc.log_with(
            NullLogTransformer(
                log_name="log_test_null",
                column_name="col_2",
                dataset_input_keys=["df_test_1", "df_test_2"],
                consume_inputs=False,
            )
        )

        log_oc.execute()

        df_log = sink_log_handle.appended

        self.assertIsNotNone(df_log)
        self.assertEqual(df_log.count(), 4)

    def test_log_step_multiple_dfs_different_order_03(self) -> None:
        sink_log_handle = TestHandle()
        source_handle = TestHandle(self.df)

        log_oc = LogOrchestrator(
            handles=[sink_log_handle],
        )

        log_oc.extract_from(
            SimpleExtractor(
                handle=source_handle,
                dataset_key="df_test_1",
            )
        )

        log_oc.log_with(
            CountLogTransformer(
                log_name="log_test_count",
                dataset_input_keys=["df_test_1"],
                consume_inputs=False,
            )
        )

        log_oc.extract_from(
            SimpleExtractor(
                handle=source_handle,
                dataset_key="df_test_2",
            )
        )

        log_oc.log_with(
            NullLogTransformer(
                log_name="log_test_null",
                column_name="col_2",
                dataset_input_keys=["df_test_1", "df_test_2"],
                consume_inputs=False,
            )
        )

        log_oc.log_with(
            CountLogTransformer(
                log_name="log_test_count",
                dataset_input_keys=["df_test_2"],
                consume_inputs=False,
            )
        )

        log_oc.execute()

        df_log = sink_log_handle.appended

        self.assertIsNotNone(df_log)
        self.assertEqual(df_log.count(), 4)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import Mock

from spetlrtools.testing import DataframeTestCase

from spetlr.etl.loaders.stream_loader import StreamLoader
from spetlr.exceptions import SparkVersionNotSupportedForSpetlrStreaming
from spetlr.filehandle import FileHandle
from spetlr.spark import Spark


# Tests are runned for spark versions lower than 10_4
# To ensure that the correct assertions are made.
@unittest.skipUnless(
    Spark.version() < Spark.DATABRICKS_RUNTIME_10_4,
    f"Sparkversion tests only applies for spark runtime versions lower than 10_4,"
    f"Your version {Spark.version()}. Skipping...",
)
class SparkVersionTests(DataframeTestCase):
    def test_01_streamloader(self):
        with self.assertRaises(SparkVersionNotSupportedForSpetlrStreaming):
            StreamLoader(
                loader=Mock(),
                options_dict={},
                await_termination=True,
                checkpoint_path="/_testpath",
            ).save(Mock())

    def test_02_filehandle(self):
        with self.assertRaises(SparkVersionNotSupportedForSpetlrStreaming):
            FileHandle(
                file_location="test", schema_location="test", data_format="test"
            ).read_stream()



from abc import abstractmethod

from pyspark.sql import DataFrame

from .types import EtlBase, dataset_group


class Extractor(EtlBase):
    """In some cases extractors depend on other extractors,
    for these situations, the datasets from previous steps
    are accessible in the variable self.previous_extractions.

    In regards to the etl step, the extractor ADDs its dataset
    to the total set of datasets.
    """

    def __init__(self, dataset_key: str = None):
        super().__init__()
        if dataset_key is None:
            dataset_key = type(self).__name__
        self.dataset_key = dataset_key
        self.previous_extractions: dataset_group = {}

    def etl(self, inputs: dataset_group) -> dataset_group:
        self.previous_extractions = inputs
        new_df = self.read()
        inputs[self.dataset_key] = new_df
        return inputs

    @abstractmethod
    def read(self) -> DataFrame:
        raise NotImplementedError()



from unittest.mock import Mock, patch

from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from spetlrtools.testing import DataframeTestCase

from spetlr.etl import Transformer
from spetlr.etl.extractors import IncrementalExtractor, SimpleExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.exceptions import MissingUpsertJoinColumns
from spetlr.orchestrators import EhToDeltaSilverOrchestrator
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaTransformer import (
    EhJsonToDeltaTransformer,
)
from spetlr.spark import Spark


class EhToDeltaSilverOrchestratorTests(DataframeTestCase):
    def test_01_default_upsert(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(IncrementalExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(SimpleLoader, "save", return_value=test_df) as p2:
                    orchestrator = EhToDeltaSilverOrchestrator(
                        dh_source=dh_source_mock,
                        dh_target=dh_target_mock,
                        upsert_join_cols=["id"],
                    )
                    orchestrator.execute()

                    p0.assert_called_once()
                    p1.assert_called_once()
                    p2.assert_called_once()

    def test_02_input_upsert(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(IncrementalExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(SimpleLoader, "save", return_value=test_df) as p2:
                    orchestrator = EhToDeltaSilverOrchestrator(
                        dh_source=dh_source_mock,
                        dh_target=dh_target_mock,
                        mode="upsert",
                        upsert_join_cols=["id"],
                    )
                    orchestrator.execute()

                    p0.assert_called_once()
                    p1.assert_called_once()
                    p2.assert_called_once()

    def test_03_upsert_missing_upsert_cols(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(IncrementalExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(SimpleLoader, "save", return_value=test_df) as p2:
                    with self.assertRaises(MissingUpsertJoinColumns) as cm:
                        orchestrator = EhToDeltaSilverOrchestrator(
                            dh_source=dh_source_mock,
                            dh_target=dh_target_mock,
                            mode="upsert",
                        )
                        orchestrator.execute()

                        p0.assert_called_once()
                        p1.assert_called_once()
                        p2.assert_called_once()
                        cm.exception = "You must specify upsert_join_cols"

    def test_04_w_transformer(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        class TestFilter(Transformer):
            def process(self, df: DataFrame) -> DataFrame:
                return df.withColumn("Test", f.lit("Test"))

        class TestOrchestrator(EhToDeltaSilverOrchestrator):
            def __init__(self):
                super().__init__(
                    dh_source=dh_source_mock,
                    dh_target=dh_target_mock,
                    upsert_join_cols=["id"],
                )
                self.filter_with(TestFilter())

        with patch.object(
            EhJsonToDeltaTransformer, "process", return_value=test_df
        ) as p1:
            with patch.object(TestFilter, "process", return_value=test_df) as p3:
                orchestrator = TestOrchestrator()
                orchestrator.execute()

                dh_source_mock.read.assert_called_once()
                p1.assert_called_once()
                dh_target_mock.upsert.assert_called_once()
                p3.assert_called_once()

    def test_05_append(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(IncrementalExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(SimpleLoader, "save", return_value=test_df) as p2:
                    orchestrator = EhToDeltaSilverOrchestrator(
                        dh_source=dh_source_mock,
                        dh_target=dh_target_mock,
                        mode="append",
                    )
                    orchestrator.execute()

                    p0.assert_called_once()
                    p1.assert_called_once()
                    p2.assert_called_once()

    def test_06_overwrite(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(SimpleExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(SimpleLoader, "save", return_value=test_df) as p2:
                    orchestrator = EhToDeltaSilverOrchestrator(
                        dh_source=dh_source_mock,
                        dh_target=dh_target_mock,
                        mode="overwrite",
                    )
                    orchestrator.execute()

                    p0.assert_called_once()
                    p1.assert_called_once()
                    p2.assert_called_once()



import json
from datetime import timedelta, timezone

from pyspark.sql.types import (
    BinaryType,
    LongType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from spetlrtools.testing import DataframeTestCase
from spetlrtools.testing.TestHandle import TestHandle
from spetlrtools.time import dt_utc

from spetlr.orchestrators.eh2bronze.EhToDeltaBronzeTransformer import (
    EhToDeltaBronzeTransformer,
)
from spetlr.spark import Spark
from spetlr.utils import DataframeCreator


class EhtoDeltaTransformerUnitTests(DataframeTestCase):
    _target_schema = StructType(
        [
            StructField("EventhubRowId", LongType(), True),
            StructField("BodyId", LongType(), True),
            StructField("Body", StringType(), True),
            StructField("EnqueuedTimestamp", TimestampType(), True),
            StructField("StreamingTime", TimestampType(), True),
            StructField("SequenceNumber", LongType(), True),
            StructField("Offset", StringType(), True),
            StructField("SystemProperties", StringType(), True),
            StructField("Properties", StringType(), True),
            StructField("pdate", TimestampType(), True),
        ]
    )

    _target_cols = [
        "EventhubRowId",
        "BodyId",
        "Body",
        "EnqueuedTimestamp",
        "StreamingTime",
        "SequenceNumber",
        "Offset",
        "SystemProperties",
        "Properties",
        "pdate",
    ]

    _capture_eventhub_output_schema = StructType(
        [
            StructField("SequenceNumber", LongType(), True),
            StructField("Offset", StringType(), True),
            StructField("SystemProperties", StringType(), True),
            StructField("Properties", StringType(), True),
            StructField("Body", BinaryType(), True),
            StructField("pdate", TimestampType(), True),
            StructField("EnqueuedTimestamp", TimestampType(), True),
        ]
    )

    def test_01(self):
        test_handle = TestHandle(
            provides=DataframeCreator.make_partial(
                self._target_schema, self._target_cols, []
            )
        )

        df_in = Spark.get().createDataFrame(
            [
                (
                    22222,
                    "OffsetTest",
                    "SystemPropertiesTest",
                    "PropertiesTest",
                    json.dumps(
                        {
                            "id": "1234",
                            "name": "John",
                        }
                    ).encode(
                        "utf-8"
                    ),  # Body
                    dt_utc(2021, 10, 31, 0, 0, 0),  # pdate
                    dt_utc(2021, 10, 31, 0, 0, 0),  # EnqueuedTimestamp
                ),
            ],
            self._capture_eventhub_output_schema,
        )

        expected = [
            (
                298587303917952901,  # EventhubRowId
                146072039196263699,  # BodyId
                json.dumps(  # Body (as string)
                    {
                        "id": "1234",
                        "name": "John",
                    }
                ),
                dt_utc(2021, 10, 31, 0, 0, 0),  # EnqueuedTimestamp
                22222,
                "OffsetTest",
                "SystemPropertiesTest",
                "PropertiesTest",
                dt_utc(2021, 10, 31, 0, 0, 0),  # pdate
            ),
        ]

        df_result = EhToDeltaBronzeTransformer(test_handle).process(df_in)

        # Assert on all column except streaming time
        self.assertDataframeMatches(
            df_result,
            [
                "EventhubRowId",
                "BodyId",
                "Body",
                "EnqueuedTimestamp",
                "SequenceNumber",
                "Offset",
                "SystemProperties",
                "Properties",
                "pdate",
            ],
            expected,
        )

        # The streaming is tested assuming that the time of the transformation
        # was less than 5 min ago
        test_time = dt_utc()
        self.assertLess(
            test_time
            - df_result.collect()[0]["StreamingTime"].replace(tzinfo=timezone.utc),
            timedelta(minutes=5),
        )



from spetlrtools.testing import DataframeTestCase

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.spark import Spark
from spetlr.testutils.stop_test_streams import stop_test_streams


class TestStopTestStreamsOnCluster(DataframeTestCase):
    """
    Tests if a test stream is stopped.

    """

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle.from_tc("MyDb").drop_cascade()

    def test_01_stop_test(self):
        tc = Configurator()
        tc.clear_all_configurations()
        tc.set_debug()
        tc.register(
            "MyDb", {"name": "TestDb{ID}", "path": "/mnt/spetlr/silver/testdb{ID}"}
        )

        db = DbHandle.from_tc("MyDb")
        db.create()

        tc.register(
            "MyTbl",
            {
                "name": "TestDb{ID}.TestTbl",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtbl",
                "format": "delta",
                "checkpoint_path": "/mnt/spetlr/silver/testdb{ID}/_checkpoint_path_tbl",
                "query_name": "testquerytbl{ID}",
            },
        )

        mirror_cp_path = "/mnt/spetlr/silver/testdb{ID}/_checkpoint_path_tblmirror"
        tc.register(
            "MyTblMirror",
            {
                "name": "TestDb{ID}.TestTblMirror",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtblmirror",
                "format": "delta",
                "checkpoint_path": mirror_cp_path,
                "await_termination": True,
                "query_name": "testquerytblmirror{ID}",
            },
        )

        dh = DeltaHandle.from_tc("MyTbl")
        Spark.get().sql(
            f"""
                CREATE TABLE {dh.get_tablename()}
                (
                id int,
                name string
                )
                LOCATION '{Configurator().get("MyTbl", "path")}'
            """
        )

        dh_mirror = DeltaHandle.from_tc("MyTblMirror")
        Spark.get().sql(
            f"""
                CREATE TABLE {dh_mirror.get_tablename()}
                (
                id int,
                name string
                )
                LOCATION '{Configurator().get("MyTblMirror", "path")}'
            """
        )
        _query_name = Configurator().get("MyTblMirror", "query_name")

        (
            Spark.get()
            .readStream.table(dh.get_tablename())
            .writeStream.trigger(processingTime="2 seconds")
            .queryName(_query_name)
            .option(
                "checkpointLocation",
                Configurator().get("MyTblMirror", "checkpoint_path"),
            )
            .toTable(dh_mirror.get_tablename())
        )

        self.assert_stream_exists(_query_name)

        stop_test_streams()

        self.assert_stream_not_exists(_query_name)

    def assert_stream_exists(self, stream_name: str) -> None:
        _test_stream_exists = False
        for stream in Spark.get().streams.active:
            if stream.name == stream_name:
                _test_stream_exists = True

        self.assertTrue(_test_stream_exists)

    def assert_stream_not_exists(self, stream_name: str) -> None:
        _test_stream_exists = False
        for stream in Spark.get().streams.active:
            if stream.name == stream_name:
                _test_stream_exists = True

        self.assertFalse(_test_stream_exists)



import unittest

from spetlr.schema_manager.spark_schema import get_schema
from spetlr.spark import Spark
from spetlr.utils import DataframeCreator


class DataframeCreatorTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        Spark.get()

        cls.schema = get_schema(
            """
            Id INTEGER,
            measured DOUBLE,
            customer STRUCT<
                name:STRING,
                address:STRING
            >,
            product_nos ARRAY<STRUCT<
                no:INTEGER,
                name:STRING
            >>
        """
        )

    def test_full_creation(self):
        df = DataframeCreator.make(
            self.schema, [(1, 3.5, ("otto", "neverland"), [(1, "book")])]
        )
        df.show()
        rows = [row.asDict(True) for row in df.collect()]
        self.assertEqual(
            [
                dict(
                    Id=1,
                    measured=3.5,
                    customer=dict(name="otto", address="neverland"),
                    product_nos=[dict(no=1, name="book")],
                )
            ],
            rows,
        )

    def test_partial_creation(self):
        df = DataframeCreator.make_partial(
            schema=self.schema,
            columns=["Id", ("customer", ["name"]), ("product_nos", ["no"])],
            data=[
                (1, ("otto",), [(1,), (2,)]),
                (2, ("max",), []),
            ],
        )
        df.show()
        rows = [row.asDict(True) for row in df.collect()]
        self.assertEqual(
            [
                dict(
                    Id=1,
                    measured=None,
                    customer=dict(name="otto", address=None),
                    product_nos=[dict(no=1, name=None), dict(no=2, name=None)],
                ),
                dict(
                    Id=2,
                    measured=None,
                    customer=dict(name="max", address=None),
                    product_nos=[],
                ),
            ],
            rows,
        )



import copy
import json
from dataclasses import asdict, dataclass, field
from typing import Dict, List, Optional, Union

from pyspark.sql.types import StructField, StructType

from spetlr import Configurator
from spetlr.deltaspec.exceptions import InvalidSpecificationError, TableSpecNotReadable
from spetlr.deltaspec.helpers import ensureStr, standard_databricks_location
from spetlr.schema_manager import SchemaManager

_DEFAULT = object()

_DEFAULT_minReaderVersion = 2
_DEFAULT_minWriterVersion = 5
_DEFAULT_blankedPropertyKeys = ["delta.columnMapping.maxColumnId"]


@dataclass
class DeltaTableSpecBase:
    """This class represents a full specification for a delta table."""

    name: Union[str, None]
    schema: StructType
    options: Dict[str, str] = field(default_factory=dict)
    partitioned_by: List[str] = field(default_factory=list)
    tblproperties: Dict[str, str] = field(default_factory=dict)
    location: Optional[str] = None
    comment: str = None

    # blanked properties will never be retained in the constructor
    blankedPropertyKeys: List[str] = field(default_factory=list)

    def __post_init__(self):
        """This method will be called automatically after instantiation and should
        not normally be called directly."""
        # The rationale here is that a name that contains a '{' will need
        # to be run through the configurator where keys are case-sensitive.
        # once a name is free of these, it may be used in comparisons to data
        # where it needs to be lower case

        self.name = ensureStr(self.name)
        self.location = ensureStr(self.location)

        if (
            (self.name and not self.name.startswith("delta.`"))  # a real name
            and self.location
            and self.location.startswith("dbfs:/user/hive/")
        ):
            # managed table locations start with this string.
            # we should treat this like an unspecified location
            self.location = None

        self.comment = ensureStr(self.comment)

        if self.name and "{" not in self.name:
            self.name = self.name.lower()

        for col in self.partitioned_by:
            if col not in self.schema.names:
                raise InvalidSpecificationError(
                    "Supply the partitioning columns in the schema."
                )

        self.location = standard_databricks_location(self.location)

        # Experiments have shown that the statement
        # ALTER TABLE she_test.tbl ALTER COLUMN a DROP NOT NULL
        # does not actually take effect on a table.
        # So we cannot work with not-nullable columns
        # in the future, if we want to generate these types
        # of alter statement, simply remove the following line.
        self.schema = self.remove_nullability(self.schema)

        # the functionality enabled by 'delta.columnMapping.mode' = 'name',
        # is needed for column manipulation and seems to go along with another
        # property "delta.columnMapping.maxColumnId": "5" which increases when
        # a column is added. it counts all the columns that were ever present
        # we should probably ignore that column in alter statements and comparisons.
        if not self.blankedPropertyKeys:
            self.blankedPropertyKeys = copy.copy(_DEFAULT_blankedPropertyKeys)
        for key in self.blankedPropertyKeys:
            if key in self.tblproperties:
                del self.tblproperties[key]

    def set_good_defaults(self):
        """
        These default values should be set on all tables
        that want to be able to drop and reorder columns.
        This function adds the following table properties unless they are already
        set to other values:
        - delta.minReaderVersion = "2"
        - delta.minWriterVersion = "5"
        - delta.columnMapping.mode = "name"
        """
        # We cannot set these as defaults in the init function because
        # when we let spark describe a table that does not have them we should not
        # just assume that they are set, or we would never apply them to a table
        # that really does not have them.

        # these two keys are special. They are set as table properties, but are
        # not read or handled as table properties.
        if "delta.minReaderVersion" not in self.tblproperties:
            self.tblproperties["delta.minReaderVersion"] = str(
                _DEFAULT_minReaderVersion
            )
        if "delta.minWriterVersion" not in self.tblproperties:
            self.tblproperties["delta.minWriterVersion"] = str(
                _DEFAULT_minWriterVersion
            )

        # This is a necessary condition for table alterations.
        # 'delta.columnMapping.mode' = 'name'
        if "delta.columnMapping.mode" in self.tblproperties:
            if self.tblproperties["delta.columnMapping.mode"] != "name":
                print(
                    f"WARNING: The table {self.name} is specified "
                    "with a property delta.columnMapping.mode != 'name'. "
                    "Expect table alteration commands to fail."
                )
        else:
            self.tblproperties["delta.columnMapping.mode"] = "name"

    def data_name(self) -> str:
        """Get a name that can be used in spark to access the underlying data."""
        c = Configurator()
        details = c.get_all_details()
        name = self.name.format(**details) if self.name else None
        location = self.location.format(**details) if self.location else None

        data_name = name or f"delta.`{location}`"
        if not data_name:
            raise TableSpecNotReadable("Name or location required to access data.")
        return data_name

    @classmethod
    def remove_nullability(self, schema: StructType) -> StructType:
        """Return a schema where the nullability of all fields is reset to default"""
        fields = []
        for f in schema.fields:
            fields.append(
                StructField(
                    name=f.name,
                    dataType=f.dataType,
                    nullable=True,
                    metadata={k: v for k, v in f.metadata.items() if v},
                )
            )
        return StructType(fields)

    # identity manipulation
    def copy(self) -> "DeltaTableSpecBase":
        """Return an independent object that compares equal to this one."""
        return copy.deepcopy(self)

    def fully_substituted(self, name=_DEFAULT) -> "DeltaTableSpecBase":
        """Return a new DeltaTableSpec
        where name and location have been completed via the Configurator."""
        parts = asdict(self)

        c = Configurator()
        details = c.get_all_details()
        parts["name"] = self.name.format(**details) if self.name else None
        parts["location"] = self.location.format(**details) if self.location else None

        # sometimes we want to override the name of the fully substituted object
        if name is not _DEFAULT:
            parts["name"] = name

        return DeltaTableSpecBase(**parts)

    def get_sql_create(self) -> str:
        """Returns a sql statement,
         that creates the table described by the current DeltaTableSpec instance.
        This method is guaranteed to be the inverse of the `.from_sql(sql)` constructor.
        """
        schema_str = SchemaManager().struct_to_sql(self.schema, formatted=True)
        self.data_name()  # raise if we cannot point to data

        sql = (
            "CREATE TABLE "
            + (self.name or f"delta.`{self.location}`")
            + f"\n(\n  {schema_str}\n)\n"
            + "USING DELTA\n"
        )

        if self.options:
            sub_parts = [
                json.dumps(k) + " = " + json.dumps(v)
                for k, v in sorted(self.options.items())
            ]
            sql += f"OPTIONS ({', '.join(sub_parts)})\n"

        if self.partitioned_by:
            sql += f"PARTITIONED BY ({', '.join(self.partitioned_by)})\n"

        if self.location:
            sql += f"LOCATION {json.dumps(self.location)}\n"

        if self.comment:
            sql += f"COMMENT {json.dumps(self.comment)}\n"

        if self.tblproperties:
            sub_parts = [
                f"  {json.dumps(k)} = {json.dumps(v)}"
                for k, v in sorted(self.tblproperties.items())
            ]
            sql += "TBLPROPERTIES (\n" + ",\n".join(sub_parts) + "\n)\n"

        return sql



import warnings
from typing import List

import pyspark.sql.functions as F
from pyspark.sql import DataFrame, Window

from spetlr.exceptions import NoTableException
from spetlr.functions import get_unique_tempview_name
from spetlr.spark import Spark


def join_time_series_dataframes(
    dfPrimary: DataFrame,
    dfSecondary: DataFrame,
    startTimeColumnName: str,
    endTimeColumnName: str,
    idColumns: List[str],
    stateColumn: str,
) -> DataFrame:
    """
    Join changing states from two dataframes acording to time as follows.

    dfPri   dfSec   df
    --              --
    |               |
    |               s1
    |       --      --
    s1      s3      s3
    |       --      --
    |               s1
    |       --      --
    --      s4      s4
    |       --      --
    |               s2
    |       --      --
    s2      s3      s3
    |       --      --
    |               s2
    |               |
    --              --
            --
            s5
            --


    """

    # Rename columns to differentiate the two dataframes
    dfPrimary_join = dfPrimary
    for column in dfPrimary_join.columns:
        dfPrimary_join = dfPrimary_join.withColumnRenamed(column, f"dfPri_{column}")

    dfSecondary_join = dfSecondary
    for column in dfSecondary_join.columns:
        dfSecondary_join = dfSecondary_join.withColumnRenamed(column, f"dfSec_{column}")

    # Create join exception
    idMatchString = " AND ".join(
        [f"dfPri_{column} = dfSec_{column}" for column in idColumns]
    )
    timeMatchString = f"( \
            (dfSec_{startTimeColumnName} >= dfPri_{startTimeColumnName} \
                AND dfSec_{startTimeColumnName} < dfPri_{endTimeColumnName}) \
            OR (dfSec_{endTimeColumnName} >= dfPri_{startTimeColumnName} \
                AND dfSec_{endTimeColumnName} < dfPri_{endTimeColumnName})\
        )"
    joinExpr = F.expr(f"{idMatchString} AND {timeMatchString}")

    # Outer join the two dataframes acording to timestamp to get all matches over time
    df_join = dfPrimary_join.join(dfSecondary_join, on=joinExpr, how="leftouter")

    # Select data after join to remove duplicated id columns
    df_join = df_join.select(
        [F.col(f"dfPri_{column}").alias(column) for column in idColumns]
        + [
            f"dfPri_{column}"
            for column in [startTimeColumnName, endTimeColumnName, stateColumn]
        ]
        + [
            f"dfSec_{column}"
            for column in [startTimeColumnName, endTimeColumnName, stateColumn]
        ]
        + [
            F.col(f"dfPri_{column}").alias(column)
            for column in dfPrimary.columns
            if column
            not in [startTimeColumnName, endTimeColumnName, stateColumn] + idColumns
        ]
    )

    # Create state id to use for groupings because value can change back to same value
    window = Window.partitionBy(idColumns).orderBy(f"dfPri_{startTimeColumnName}")
    df_join = df_join.withColumn(
        "dfPri_StateGroupId",
        F.sum(
            (
                F.col(f"dfPri_{stateColumn}")
                != F.lag(F.col(f"dfPri_{stateColumn}")).over(window)
            ).cast("int")
        ).over(window),
    )
    df_join = df_join.fillna({"dfPri_StateGroupId": 0})

    # Fix start and end timestamps when multiple states in secondary dataframe
    # match primary dataframe
    window = Window.partitionBy(idColumns + ["dfPri_StateGroupId"]).orderBy(
        f"dfPri_{startTimeColumnName}", f"dfSec_{startTimeColumnName}"
    )
    # endtime
    df_join = df_join.withColumn(
        f"dfPri_{endTimeColumnName}",
        F.when(
            (
                F.col(f"dfPri_{endTimeColumnName}")
                == F.lead(F.col(f"dfPri_{endTimeColumnName}")).over(window)
            ),
            F.col(f"dfSec_{endTimeColumnName}"),
        ).otherwise(F.col(f"dfPri_{endTimeColumnName}")),
    )
    # start time
    df_join = df_join.withColumn(
        f"dfPri_{startTimeColumnName}",
        F.when(
            (
                F.col(f"dfPri_{startTimeColumnName}")
                == F.lag(F.col(f"dfPri_{startTimeColumnName}")).over(window)
            ),
            F.lag(F.col(f"dfSec_{endTimeColumnName}")).over(window),
        ).otherwise(F.col(f"dfPri_{startTimeColumnName}")),
    )

    # Generating three time segements:
    #   - One with the state from the primary dataframe
    #   - One with the state from the secondary dataframe
    #   - One with the rest of primary dataframe
    # Correcting the end timestamps according to if there is
    # a matching element in secondary dataframe
    df_join = df_join.withColumn(
        "StartTimeDfPri1", F.col(f"dfPri_{startTimeColumnName}")
    )
    df_join = df_join.withColumn(
        "EndTimeDfPri1",
        F.when(
            F.col(f"dfSec_{stateColumn}").isNotNull(),
            F.col(f"dfSec_{startTimeColumnName}"),
        ).otherwise(F.col(f"dfPri_{endTimeColumnName}")),
    )

    df_join = df_join.withColumn(
        "StartTimeDfSec", F.col(f"dfSec_{startTimeColumnName}")
    )
    df_join = df_join.withColumn("EndTimeDfSec", F.col(f"dfSec_{endTimeColumnName}"))

    df_join = df_join.withColumn("StartTimeDfPri2", F.col(f"dfSec_{endTimeColumnName}"))

    df_join = df_join.withColumn(
        "EndTimeDfPri2",
        F.when(
            F.col(f"dfSec_{stateColumn}").isNotNull(),
            F.col(f"dfPri_{endTimeColumnName}"),
        ).otherwise(None),
    )

    # Collecting segments of timestamps into three lists
    df_join = (
        df_join.withColumn(
            "DfPri1",
            F.when(
                F.col("StartTimeDfPri1") < F.col("EndTimeDfPri1"),
                F.array(F.col("StartTimeDfPri1"), F.col("EndTimeDfPri1")),
            ).otherwise(F.array(F.lit(None), F.lit(None))),
        )
        .withColumn(
            "DfSec",
            F.when(
                F.col("StartTimeDfSec") < F.col("EndTimeDfSec"),
                F.array(F.col("StartTimeDfSec"), F.col("EndTimeDfSec")),
            ).otherwise(F.array(F.lit(None), F.lit(None))),
        )
        .withColumn(
            "DfPri2",
            F.when(
                F.col("StartTimeDfPri2") < F.col("EndTimeDfPri2"),
                F.array(F.col("StartTimeDfPri2"), F.col("EndTimeDfPri2")),
            ).otherwise(F.array(F.lit(None), F.lit(None))),
        )
    )

    # Unpivot dataframe to move timestamp segemnts from columns to rows
    unpivot_columnn_select = (
        idColumns
        + [
            f"{column}"
            for column in dfPrimary.columns
            if column
            not in [startTimeColumnName, endTimeColumnName, stateColumn] + idColumns
        ]
        + [
            f"dfPri_{stateColumn}",
            f"dfSec_{stateColumn}",
            "stack(3, 'DfPri1', DfPri1, 'DfSec', DfSec, 'DfPri2', DfPri2) "
            "as (type, data)",
        ]
    )
    df_join_unpivot = df_join.selectExpr(*unpivot_columnn_select)

    # Extract two timestamps from unpivot data column
    df_join_unpivot = (
        df_join_unpivot.withColumn(startTimeColumnName, F.col("data")[0])
        .withColumn(endTimeColumnName, F.col("data")[1])
        .drop("data")
    )

    # Remove rows where start time and end time is null
    df_join_unpivot = df_join_unpivot.where(
        F.col(startTimeColumnName).isNotNull() | F.col(endTimeColumnName).isNotNull()
    )

    # Define stateColumn from either df primary or df seconday
    df_join_unpivot = df_join_unpivot.withColumn(
        "tempStateColumn",
        F.when(F.col("type") == "DfPri1", F.col(f"dfPri_{stateColumn}"))
        .when(F.col("type") == "DfSec", F.col(f"dfSec_{stateColumn}"))
        .when(F.col("type") == "DfPri2", F.col(f"dfPri_{stateColumn}"))
        .otherwise(None),
    )
    df_join_unpivot = df_join_unpivot.drop(
        f"dfPri_{stateColumn}", f"dfSec_{stateColumn}", "type"
    )
    df_join_unpivot = df_join_unpivot.withColumnRenamed("tempStateColumn", stateColumn)

    df_join_unpivot = df_join_unpivot.dropDuplicates(
        subset=[startTimeColumnName, endTimeColumnName]
    )

    return df_join_unpivot


def merge_df_into_target(
    df: DataFrame,
    table_name: str,
    database_name: str,
    join_cols: List[str],
    table_format: str = "Delta",
) -> None:
    """

    Merges a databricks dataframe into a target database table

    :param df: The dataframe
    :param table_name: The name of the table which the dataframe (p1)
        should be merged into
    :param database_name: The database name associated with the table
    :param join_cols: A list of strings which tells which columns to join on
    :param table_format: The format of the table

    """

    target_table_name = str(database_name + "." + table_name)

    # Check if table exists
    if not Spark.get()._jsparkSession.catalog().tableExists(target_table_name):
        raise NoTableException(f"The table {target_table_name} not found.")

    # check null keys in our dataframe.
    any_null_keys = len(
        df.filter(" OR ".join(f"({col} is NULL)" for col in join_cols)).take(1)
    )

    if any_null_keys:
        warnings.warn(
            "Null keys found in input dataframe. Rows will be discarded before load."
        )
        df = df.filter(" AND ".join(f"({col} is NOT NULL)" for col in join_cols))

    df_target = Spark.get().table(target_table_name)

    # If the target is empty, always do faster full load
    if len(df_target.take(1)) == 0:
        return (
            df.write.format(table_format)
            .mode("overwrite")
            .saveAsTable(target_table_name)
        )  # Consider use .save instead

    # Find records that need to be updated in the target (happens seldom)

    # Define the column to be used for checking for new rows
    # Checking the null-ness of one right row is sufficient to mark the row as new,
    # since null keys are disallowed.
    key_col = join_cols[-1]

    # Write the filtering string including casting maps to strings
    filter_string = f"{key_col} IS NULL"
    for col, col_typ in df.dtypes:
        if col not in join_cols:
            if col_typ == "map":
                filter_string += (
                    f" OR (CAST(a.{col} AS varchar) <> CAST(b.{col} AS varchar))"
                )
            else:
                filter_string += f" OR (a.{col} <> b.{col})"
                filter_string += f" OR (a.{col} IS NULL AND b.{col} IS NOT NULL)"
                filter_string += f" OR (a.{col} IS NOT NULL AND b.{col} IS NULL)"

    df = (
        df.alias("a")
        .join(
            df_target.alias("b").select(
                "b.*", F.col(f"b.{key_col}").alias(f"{key_col}_copy")
            ),
            on=join_cols,
            how="left",
        )
        .filter(filter_string)
        .withColumn(
            "is_new",
            F.when(F.col(f"{key_col}_copy").isNull(), True).otherwise(False),
        )
        .select("a.*", "is_new")
        .cache()
    )

    # If merge is not required, the data can just be appended
    merge_required = len(df.filter(~F.col("is_new")).take(1)) > 0
    df = df.drop("is_new")

    if not merge_required:
        return (
            df.write.format(table_format).mode("append").saveAsTable(target_table_name)
        )  # Consider use .save instead

    temp_view_name = get_unique_tempview_name()
    df.createOrReplaceTempView(temp_view_name)

    merge_sql_statement = f"""
        MERGE INTO {target_table_name} AS target
        USING {temp_view_name} as source
        ON
        {" AND ".join(f"(source.{col} = target.{col})" for col in join_cols)}
        WHEN MATCHED
        THEN UPDATE -- update existing records
        SET
        *
        WHEN NOT MATCHED
        THEN INSERT -- insert new records
        *
        """
    Spark.get().sql(merge_sql_statement)


def concat_dfs(dfs: List[DataFrame]):
    """
    concat_dfs append dataframes on eachother and also append columns.
         If columns has the exact same name, they will be converted to one column.

    NB: The transformation does not perform dataframe joins.

    :param dfs: A list of spark dataframes

    returns a dataframe

    """

    # Check that the list of dataframe and each dataframe is not Nonetypes
    if dfs is None:
        raise NoTableException("The table list of tables are None")

    for i, df in enumerate(dfs):
        if df is None:
            raise NoTableException(f"The {i}th table dataframe in the list is None.")

    all_cols = []
    dfs_updated = dfs

    # Get all columns across all tables
    for df in dfs:
        all_cols = all_cols + df.columns

    # Sort the columns
    all_cols = sorted(list(set(all_cols)))

    # For each dataframe: Add the columns from all_cols if the dataframe does not has it
    for i, df in enumerate(dfs):
        for col in all_cols:
            if col not in df.columns:
                dfs_updated[i] = dfs_updated[i].withColumn(col, F.lit(None))

        # Select columns in correct order
        dfs_updated[i] = dfs_updated[i].select(all_cols)

    # Union the dataframes
    for i, df in enumerate(dfs):
        if i == 0:
            result = df
        else:
            result = result.unionByName(df)

    return result



from more_itertools import peekable

# This type should express that it is a peekable iterator
# that returns TokenLists. It seems that this is not easily
# expressed. Here is what I had tried:
# peekable[TokenList]
_PeekableTokenList = peekable
